@@ -3,11 +3,12 @@ import datetime
 import logging
 import pathlib
 import re
+import threading
 import time
 import xml.etree.ElementTree as ET
 from io import SEEK_CUR, SEEK_END, SEEK_SET, BufferedReader, BytesIO
 from pathlib import Path
-from typing import Dict, Optional, Union
+from typing import Dict, List, Optional, Union
 from urllib.parse import urlencode
 
 import requests
@@ -31,6 +32,7 @@ from .datastructures import (
     PolicyList,
     PolicyUser,
     RMCClient,
+    ServerSettings,
     ShareActivity,
     SharedType,
     SortBy,
@@ -50,6 +52,39 @@ def str_to_bool(value):
 log = logging.getLogger(__name__)
 
 
+class Progress:
+    """
+    Way to track progress of uploads/downloads.
+
+    Either use this object in another thread or
+    override update() to get progress updates.
+    """
+
+    def __init__(self) -> None:
+        self._completed_bytes = 0
+        self._total_bytes = 0
+        self._lock = threading.Lock()
+
+    """
+    Progress callback of uploads/downloads
+    """
+
+    def update(
+        self, completed_bytes: int, total_bytes: int, chunk_complete: bool
+    ) -> None:
+        with self._lock:
+            self._completed_bytes = completed_bytes
+            self._total_bytes = total_bytes
+
+    def completed_bytes(self) -> int:
+        with self._lock:
+            return self._completed_bytes
+
+    def total_bytes(self) -> int:
+        with self._lock:
+            return self._total_bytes
+
+
 class FCServer:
     """
     FileCloud Server API
@@ -496,7 +531,11 @@ class FCServer:
         raise TimeoutError(f"File {path} not removed after {maxwaits} seconds")
 
     def downloadfile_no_retry(
-        self, path: str, dstPath: Union[pathlib.Path, str], redirect: bool = True
+        self,
+        path: str,
+        dstPath: Union[pathlib.Path, str],
+        redirect: bool = True,
+        progress: Optional[Progress] = None,
     ) -> None:
         """
         Download file at 'path' to local 'dstPath'
@@ -511,23 +550,32 @@ class FCServer:
             stream=True,
         ) as resp:
             resp.raise_for_status()
+            content_length = int(resp.headers.get("Content-Length", "-1"))
+            completed_bytes = 0
             with open(dstPath, "wb") as dstF:
                 for chunk in resp.iter_content(128 * 1024):
+                    completed_bytes += len(chunk)
                     dstF.write(chunk)
+                    if progress is not None:
+                        progress.update(completed_bytes, content_length, False)
 
     def downloadfile(
-        self, path: str, dstPath: Union[pathlib.Path, str], redirect: bool = True
+        self,
+        path: str,
+        dstPath: Union[pathlib.Path, str],
+        redirect: bool = True,
+        progress: Optional[Progress] = None,
     ) -> None:
         """
         Download file at 'path' to local 'dstPath'. Retries.
         """
         if self.retries is None:
-            return self.downloadfile_no_retry(path, dstPath, redirect)
+            return self.downloadfile_no_retry(path, dstPath, redirect, progress)
 
         retries = self.retries
         while True:
             try:
-                self.downloadfile_no_retry(path, dstPath, redirect)
+                self.downloadfile_no_retry(path, dstPath, redirect, progress)
                 return
             except:
                 retries = retries.increment()
@@ -568,29 +616,45 @@ class FCServer:
         data: bytes,
         serverpath: str,
         datemodified: datetime.datetime = datetime.datetime.now(),
+        nofileoverwrite: Optional[bool] = False,
+        iflastmodified: Optional[datetime.datetime] = None,
+        progress: Optional[Progress] = None,
     ) -> None:
         """
         Upload bytes 'data' to server at 'serverpath'.
         """
-        self.upload(BufferedReader(BytesIO(data)), serverpath, datemodified)  # type: ignore
+        self.upload(BufferedReader(BytesIO(data)), serverpath, datemodified, nofileoverwrite=nofileoverwrite, iflastmodified=iflastmodified, progress=progress)  # type: ignore
 
     def upload_str(
         self,
         data: str,
         serverpath: str,
         datemodified: datetime.datetime = datetime.datetime.now(),
+        nofileoverwrite: Optional[bool] = False,
+        iflastmodified: Optional[datetime.datetime] = None,
+        progress: Optional[Progress] = None,
     ) -> None:
         """
         Upload str 'data' UTF-8 encoded to server at 'serverpath'.
         """
-        self.upload_bytes(data.encode("utf-8"), serverpath, datemodified)
+        self.upload_bytes(
+            data.encode("utf-8"),
+            serverpath,
+            datemodified,
+            nofileoverwrite=nofileoverwrite,
+            iflastmodified=iflastmodified,
+            progress=progress,
+        )
 
     def upload_file(
         self,
         localpath: pathlib.Path,
         serverpath: str,
         datemodified: datetime.datetime = datetime.datetime.now(),
+        nofileoverwrite: Optional[bool] = False,
+        iflastmodified: Optional[datetime.datetime] = None,
         adminproxyuserid: Optional[str] = None,
+        progress: Optional[Progress] = None,
     ) -> None:
         """
         Upload file at 'localpath' to server at 'serverpath'.
@@ -600,7 +664,10 @@ class FCServer:
                 uploadf,
                 serverpath,
                 datemodified,
+                nofileoverwrite,
+                iflastmodified,
                 adminproxyuserid=adminproxyuserid,
+                progress=progress,
             )
 
     def _serverdatetime(self, dt: datetime.datetime):
@@ -618,7 +685,10 @@ class FCServer:
         uploadf: BufferedReader,
         serverpath: str,
         datemodified: datetime.datetime,
+        nofileoverwrite: Optional[bool] = False,
+        iflastmodified: Optional[datetime.datetime] = None,
         adminproxyuserid: Optional[str] = None,
+        progress: Optional[Progress] = None,
     ) -> None:
         """
         Upload seekable stream at uploadf to server at 'serverpath'
@@ -681,6 +751,8 @@ class FCServer:
                     size = min(size, max_read)
                 data = super().read(size)
                 self.pos += len(data)
+                if progress is not None:
+                    progress.update(self.pos, data_size, False)
                 return data
 
             def __len__(self) -> int:
@@ -744,6 +816,13 @@ class FCServer:
                 "date": self._serverdatetime(datemodified),
                 "adminproxyuserid": adminproxyuserid,
             }
+
+            if nofileoverwrite is not None:
+                params["nofileoverwrite"] = 1 if nofileoverwrite else 0
+
+            if iflastmodified is not None:
+                params["iflastmodified"] = str(int(iflastmodified.timestamp()))
+
             params_str = urlencode(params)
 
             if params_str.find("%2FSHARED%2F%21"):
@@ -786,6 +865,12 @@ class FCServer:
                 "adminproxyuserid": adminproxyuserid,
             }
 
+            if nofileoverwrite is not None:
+                params["nofileoverwrite"] = 1 if nofileoverwrite else 0
+
+            if iflastmodified is not None:
+                params["iflastmodified"] = str(int(iflastmodified.timestamp()))
+
             if data_size is not None:
                 params["filesize"] = data_size
 
@@ -811,6 +896,9 @@ class FCServer:
 
             pos += curr_slice_size
 
+            if progress is not None:
+                progress.update(pos, data_size, True)
+
     def share(self, path: str, adminproxyuserid: str = "") -> FCShare:
         """
         Share 'path'
@@ -843,6 +931,34 @@ class FCServer:
             str_to_bool(resp.findtext("./share/allowpublicuploadonly", "")),
         )
 
+    def quickshare(self, sharelocation: str, adminproxyuserid: str = "") -> FCShare:
+        """
+        Quick Share 'sharelocation'
+        """
+        resp = self._api_call(
+            "/core/quickshare",
+            {"sharelocation": sharelocation, "adminproxyuserid": adminproxyuserid},
+        )
+
+        shareid = resp.findtext("./share/shareid", "")
+
+        if not shareid:
+            msg = resp.findtext("./meta/message", "")
+            if msg:
+                raise ServerError("", msg)
+            else:
+                raise ServerError("", "No shareid in response")
+
+        return FCShare(
+            shareid,
+            resp.findtext("./share/sharename", ""),
+            resp.findtext("./share/sharelocation", ""),
+            str_to_bool(resp.findtext("./share/allowpublicaccess", "")),
+            str_to_bool(resp.findtext("./share/allowpublicupload", "")),
+            str_to_bool(resp.findtext("./share/allowpublicviewonly", "")),
+            str_to_bool(resp.findtext("./share/allowpublicuploadonly", "")),
+        )
+
     def getshareforpath(self, path: str, adminproxyuserid: str = "") -> FCShare:
         """
         Share 'path'
@@ -911,6 +1027,15 @@ class FCServer:
 
         self._raise_exception_from_command(resp)
 
+    def addgrouptoshare(self, share: FCShare, groupid: str) -> None:
+        """
+        Allow group access to share
+        """
+        resp = self._api_call(
+            "/core/addgrouptoshare", {"shareid": share.shareid, "groupid": groupid}
+        )
+        self._raise_exception_from_command(resp)
+
     def createfolder(
         self,
         path: str,
@@ -984,6 +1109,35 @@ class FCServer:
         )
         self._raise_exception_from_command(resp)
 
+    def setgroupaccessforshare(
+        self,
+        share: FCShare,
+        groupid: str,
+        allowwrite: bool,
+        allowdownload: bool,
+        allowshare: bool,
+        allowsync: bool,
+        disallowdelete: bool,
+        adminproxyuserid: Optional[str] = None,
+    ) -> None:
+        """
+        Set group permissions for share
+        """
+        resp = self._api_call(
+            "/core/setgroupaccessforshare",
+            {
+                "shareid": share.shareid,
+                "groupid": groupid,
+                "write": "true" if allowwrite else "false",
+                "download": "true" if allowdownload else "false",
+                "share": "true" if allowshare else "false",
+                "sync": "true" if allowsync else "false",
+                "disallowdelete": "true" if disallowdelete else "false",
+                "adminproxyuserid": adminproxyuserid if adminproxyuserid else "",
+            },
+        )
+        self._raise_exception_from_command(resp)
+
     def getusersforshare(self, share: FCShare) -> list[FCShareUser]:
         """
         Returns a list of users that are added explicitly to the share
@@ -1507,6 +1661,30 @@ class FCServer:
 
         self._raise_exception_from_command(resp)
 
+    def admin_get_config_settings(self, config: List[str]) -> ServerSettings:
+        """
+        Retrieve Filecloud configuration settings. The config list should
+        contain FC configuration keys.
+        Args:
+            config: List containing FC config keys
+        """
+        try:
+            count = len(config)
+        except (ValueError, TypeError):
+            count = 0
+
+        config_opts = {}
+        for i in range(count):
+            param_key = f"param{i}"
+            config_opts[param_key] = config[i]  # type:ignore
+
+        config_opts["count"] = str(len(config_opts))
+
+        resp = self._api_call("/admin/getconfigsetting", config_opts)
+
+        settings = ServerSettings(resp)
+        return settings
+
     def set_config_setting(self, config_name: str, config_val: str) -> None:
         """
         Sets a server config setting via admin
